/*
 SOS Project Classes.

 Copyright (C) 2006 SOS Project <http://sos.project.free.fr>
 SOS Project is :
 Denis BURGARELLA @ OAMP <http://www.lam.oamp.fr/denis>
 Camille AMBROSI @ CMI <http://camille.ambrosi.free.fr>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package simulation.components;

import kernel.Maths;
import kernel.Kernel;
import simulation.SimulationSpec;

public abstract class ImageComponent
{

	/* position du composant */
	protected double			xCoord			= 0.0;

	protected double			yCoord			= 0.0;

	/* informations sur la simulation */
	protected SimulationSpec	spec			= null;

	/* le flux total en Jy */
	protected double			dblTotalFlux	= 0.0;	// le flux total initial
														// demander

	protected double			dblRealFlux		= 0.0;	// le flux total
														// reelement obtenu

	protected double			dblRealElect	= 0.0;	// le nombre total
														// d'electron reelement
														// obtenu

	/* les couleurs à calculer */
	// protected double[] resRaies = null;
	/* le flux de l'objet */
	// protected double dblFluxR = 0.0;
	// protected double dblFluxG = 0.0;
	// protected double dblFluxB = 0.0;

	/**
	 * Constructeur générique d'un composant.
	 * 
	 * @param spec Les spécifications de la simulation.
	 * @param xCoord La position X du composant.
	 * @param yCoord La position Y du composant.
	 */
	protected ImageComponent(final SimulationSpec spec, final double xCoord,
			final double yCoord, final double dblTotalFlux)
	{
		try {
			/* position */
			this.xCoord = xCoord;
			this.yCoord = yCoord;
			/* flux */
			this.dblTotalFlux = dblTotalFlux;
			/* spécifications */
			this.spec = spec;
			/* filtres */
			// resRaies = new double[3];
		} catch (Exception ex) {
			Kernel.fireError(this.getClass().getName(), "ImageComponent()", ex);
		}
	}

	/**
	 * Génère le composant dans une matrice de pixels.
	 */
	public final void generate()
	{
		try {

			/* optimisation */
			final int iSizeMax = spec.GENERATION_OBJECTS_SIZE;

			/* positions maximales et minimales */
			final int xMax = iSizeMax * 2 + 1;
			final int yMax = iSizeMax * 2 + 1;

			/* retrouve le point initial */
			final int xPixFromOrigin = (int) xCoord;
			final int yPixFromOrigin = (int) yCoord;

			/* positions absolues de la matrice temporaire */
			final int xAbsStart = xPixFromOrigin - iSizeMax;
			final int yAbsStart = yPixFromOrigin - iSizeMax;

			/* matrice temporaire */
			double[][] tmp = new double[xMax][yMax];

			/* distances actuelles du centre */
			double xDistFromCenter;
			double yDistFromCenter;

			double dblTotalVal = 0.0;
			double dblVal;

			for (int j = 0; j < yMax; j++) {
				for (int i = 0; i < xMax; i++) {
					/* calcule la distance du centre de l'objet */
					xDistFromCenter = xAbsStart + i + 0.5 - xCoord;
					yDistFromCenter = yAbsStart + j + 0.5 - yCoord;
					/* calcule l'intensite arbitraire */
					dblVal = calculateIntensity(xDistFromCenter,
							yDistFromCenter, spec.getFilterMono());
					/* memorise cette donnee */
					dblTotalVal += dblVal;
					tmp[i][j] = dblVal;
				}
			}

			dblRealFlux = 0.0;
			dblRealElect = 0.0;

			/* on a l'intensite totale, on la ramene a la valeur souhaitee */
			if (dblTotalVal > 0) {
				/* on calcule le facteur */
				double dblFactor = dblTotalFlux / dblTotalVal;
				/* on stocke cette donnee dans la matrice principale */
				for (int j = 0; j < yMax; j++) {
					for (int i = 0; i < xMax; i++) {
						/*
						 * on verifie que l'on reste bien dans la matrice
						 * principale
						 */
						if ((xAbsStart + i > -1)
								&& (xAbsStart + i < spec.PARAM_IMAGE_SIZE_X)
								&& (yAbsStart + j > -1)
								&& (yAbsStart + j < spec.PARAM_IMAGE_SIZE_Y)) {
							/* flux reel */
							dblVal = tmp[i][j] * dblFactor;
							dblRealFlux += dblVal;
							/* nombre d'electron */
							dblVal = Maths.fromJyToElec(dblVal, spec
									.getFilterMono().lambdaCenterInMeter(),
									spec.getFilterMono().dlambdaInMeter(),
									spec.PARAM_OBSERVATION_TIME,
									spec.PARAM_TELESCOPE_DIAMETER);
							dblRealElect += dblVal;
							/*
							 * on ajoute le flux que l'on convertie de Jy en
							 * nombre d'electron
							 */
							spec.getMatrix().addObjectsElectrons(xAbsStart + i,
									yAbsStart + j, dblVal);
						}
					}
				}
			}

			// System.out.println("tot = "+dblTotalFlux);
			// System.out.println("real = "+dblRealFlux);

			// System.out.println("elec = "+dblRealElect);
			// System.out.println("elec = "+Functions.fromJyToElec2(dblRealFlux,
			// spec.getFilterMono(), spec.getObservationTime(),
			// spec.getTelescDiameter()));

		} catch (Exception ex) {
			Kernel.fireError(this.getClass().getName(), "generate()", ex);
		}
	}

	public final double getRefTotalFluxMONO()
	{
		return dblTotalFlux;
	}

	public final double getRealTotalFluxMONO()
	{
		return dblRealFlux;
	}

	public final double getRealTotalElectNbMONO()
	{
		return dblRealElect;
	}

	// public final double getFluxR() {return dblFluxR * spec.getPixelField() *
	// spec.getPixelField();}
	// public final double getFluxG() {return dblFluxG * spec.getPixelField() *
	// spec.getPixelField();}
	// public final double getFluxB() {return dblFluxB * spec.getPixelField() *
	// spec.getPixelField();}

	public final double getPosX()
	{
		return xCoord;
	}

	public final double getPosY()
	{
		return yCoord;
	}

	/**
	 * Calcule l'intensité d'une galaxie en fonction de la coordonnée.
	 * 
	 * @param x La coordonnée X du point.
	 * @param y La coordonnée Y du point.
	 * @return L'intensité.
	 */
	protected abstract double calculateIntensity(final double x,
			final double y, final Filter filter);

}
