/*
 SOS Project Classes.

 Copyright (C) 2006 SOS Project <http://sos.project.free.fr>
 SOS Project is :
 Denis BURGARELLA @ OAMP <http://www.lam.oamp.fr/denis>
 Camille AMBROSI @ CMI <http://camille.ambrosi.free.fr>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package simulation.components;

import java.awt.image.BufferedImage;

import kernel.Kernel;
import simulation.SimulationSpec;

/**
 * 
 * @author camille AMBROSI
 * 
 * Une matrice de pixel contient toutes les donnees de l'image. La valeur de
 * chaque pixel est le nombre d'electron par seconde par pixel
 * [elec.s(-1).px(-1)]. Cette valeur peut facilement etre convertie en jansky
 * par steradian [Jy.sterad(-1)] grace a la formule suivante : <code>
 *   flux [Jy.px(-1)] 		= flux [elec.s(-1).px(-1)] * 
 *   flux [Jy.arcsec(-1)]	= flux [Jy.px(-1)] / pixres [arcsec.px[-1]]
 *   flux [Jy.sterad(-1)]	= flux [Jy.arcsec(-1)] * PI / 3600^2 / 180^2
 * </code>
 * 
 */
public class ImageMatrix
{

	public static final int	VALUES				= 3;

	public static final int	VALUE_NORMALIZED	= 0;

	public static final int	VALUE_OBJECTS		= 1;

	public static final int	VALUE_NOISES		= 2;

	/* les spécification de la simulation */
	private SimulationSpec	simulSpec			= null;

	/* la matrice de pixels */
	private double[][][]	data				= null;

	/* mémorise si la matrice est normalisé */
	private boolean			blIsNormalized		= false;

	/* les dimensions de la matrice */
	private int				iTailleX			= 0;

	private int				iTailleY			= 0;

	/* l'image */
	private BufferedImage	img					= null;

	public ImageMatrix(final SimulationSpec simulSpec)
	{
		try {
			this.simulSpec = simulSpec;
			/* nouvelle matrice */
			if ((simulSpec.PARAM_IMAGE_SIZE_X > 0)
					&& (simulSpec.PARAM_IMAGE_SIZE_Y > 0)) {
				this.iTailleX = simulSpec.PARAM_IMAGE_SIZE_X;
				this.iTailleY = simulSpec.PARAM_IMAGE_SIZE_Y;
				this.data = new double[iTailleX][iTailleY][VALUES];
			}
		} catch (Exception ex) {
			Kernel.fireError(this.getClass().getName(), "ImageMatrix()", ex);
		}
	}

	public final void resetMatrix()
	{
		try {
			simulSpec.fireLog(" |  | ");
			simulSpec.fireLog(" |  | initialisation de la matrice...");
			/* initialise les valeurs */
			blIsNormalized = false;
			img = null;
			/* on verifie les parametres */
			if ((simulSpec.PARAM_IMAGE_SIZE_X < 1)
					|| (simulSpec.PARAM_IMAGE_SIZE_Y < 1)) {
				simulSpec
						.fireLog(" |  |  | impossible de créer la matrice : paramêtres incorrects.");
				simulSpec.fireLog(" |  |  \\_erreur.");
				return;
			}
			/* on regarde si on doit créé une nouvelle matrice ou juste la vider */
			if ((iTailleX != simulSpec.PARAM_IMAGE_SIZE_X)
					|| (iTailleY != simulSpec.PARAM_IMAGE_SIZE_Y)
					|| (data == null)) {
				simulSpec.fireLog(" |  |  | nouvelle matrice allouée.");
				/* on memorise la nouvelle taille */
				this.iTailleX = simulSpec.PARAM_IMAGE_SIZE_X;
				this.iTailleY = simulSpec.PARAM_IMAGE_SIZE_Y;
				/* on efface l'ancienne matrice */
				if (data != null) {
					data = null;
					System.gc();
				}
				/* on construit notre matrice de pixels */
				data = new double[iTailleX][iTailleY][VALUES];
			} else {
				/* on initialise les pixels */
				for (int i = 0; i < iTailleX; i++) {
					for (int j = 0; j < iTailleY; j++) {
						for (int k = 0; k < VALUES; k++) {
							data[i][j][k] = 0.0;
						}
					}
				}
				simulSpec.fireLog(" |  |  | matrice vidée.");
			}
			simulSpec.fireLog(" |  |  \\_ok.");
			return;
		} catch (Exception ex) {
			Kernel.fireError(this.getClass().getName(), "initMatrix()", ex);
		}
		simulSpec.fireLog(" |  |  \\_erreur.");
	}

	public final void normalize()
	{
		try {
			double dlbIntensMax = 0.0;
			double dblFactor = -1.0;
			this.blIsNormalized = false;

			simulSpec.fireLog(" |  | ");
			simulSpec.fireLog(" |  | normalisation de la matrice...");

			/* on recherche la valeur maximale */
			for (int i = 0; i < iTailleX; i++) {
				for (int j = 0; j < iTailleY; j++) {
					if ((data[i][j][VALUE_OBJECTS] + data[i][j][VALUE_NOISES]) > dlbIntensMax) dlbIntensMax = (data[i][j][VALUE_OBJECTS] + data[i][j][VALUE_NOISES]);
				}
			}
			dlbIntensMax = Math
					.pow(dlbIntensMax, simulSpec.NORMALISATION_POWER);
			simulSpec.fireLog(" |  |  | intensité maximale : " + dlbIntensMax);

			/* on verifie que la matrice ne soit pas vide */
			if (dlbIntensMax <= 0) {
				simulSpec.fireLog(" |  |  | impossible de normaliser");
				simulSpec.fireLog(" |  |  \\_fin.");
				return;
			}

			/* on calcule le facteur */
			dblFactor = (255.0 + (double) (simulSpec.NORMALISATION_VALUE))
					/ dlbIntensMax;
			simulSpec.fireLog(" |  |  | facteur de normalisation : "
					+ dblFactor);
			simulSpec.fireLog(" |  |  | normalisation des pixels...");

			/* on normalise les pixels */
			for (int i = 0; i < iTailleX; i++) {
				for (int j = 0; j < iTailleY; j++) {
					/* on applique la normalisaton */
					if ((data[i][j][VALUE_OBJECTS] + data[i][j][VALUE_NOISES]) <= 0.0) {
						data[i][j][VALUE_NORMALIZED] = 0.0;
					} else {
						data[i][j][VALUE_NORMALIZED] = Math
								.pow(
										(data[i][j][VALUE_OBJECTS] + data[i][j][VALUE_NOISES]),
										simulSpec.NORMALISATION_POWER)
								* dblFactor + 0.49;
					}
				}
			}

			/* on memorise que l'on est normalise */
			blIsNormalized = true;
			simulSpec.fireLog(" |  |  \\_ok.");
			return;

		} catch (Exception ex) {
			Kernel.fireError(this.getClass().getName(), "normalize()", ex);
		}
		simulSpec.fireLog(" |  |  \\_erreur.");
	}

	public final void createImage()
	{
		try {
			simulSpec.fireLog(" |  | ");
			simulSpec.fireLog(" |  | création de l'image...");

			if (blIsNormalized == false) {

				/* la matrice n'est pas normalisee, impossible de creer l'image */
				simulSpec
						.fireLog(" |  |  | la matrice n'est pas normalisee, impossible de creer l'image.");
				simulSpec.fireLog(" |  |  \\_fin.");
				img = null;

			} else {

				simulSpec.fireLog(" |  |  | taille de l'image : " + iTailleX
						+ "px X " + iTailleY + "px.");
				simulSpec.fireLog(" |  |  | conversion des pixels...");

				/* on cre une image vide */
				img = new BufferedImage(iTailleX, iTailleY,
						BufferedImage.TYPE_INT_RGB);

				/* on copie les pixels uns a uns */
				for (int i = 0; i < iTailleX; i++) {
					for (int j = 0; j < iTailleY; j++) {
						img.setRGB(i, j, getRGBColorValue(i, j));
					}
				}

				simulSpec.fireLog(" |  |  \\_ok.");
			}

			return;
		} catch (Exception ex) {
			Kernel.fireError(this.getClass().getName(), "createImage()", ex);
		}
		simulSpec.fireLog(" |  |  \\_erreur.");
	}

	public final int getNormalizedValue(final int i, final int j)
	{
		int val = (int) data[i][j][VALUE_NORMALIZED];
		if (val > 255) val = 255;
		return val;
	}

	public final int getRGBColorValue(final int i, final int j)
	{
		int val = getNormalizedValue(i, j);
		return ((255 & 0xFF) << 24) | ((val & 0xFF) << 16)
				| ((val & 0xFF) << 8) | ((val & 0xFF) << 0);
	}

	public final void setNoisesElectrons(final int i, final int j,
			final double dblValue)
	{
		data[i][j][VALUE_NOISES] = dblValue;
	}

	public final double getNoisesElectrons(final int i, final int j)
	{
		return data[i][j][VALUE_NOISES];
	}

	public final void addObjectsElectrons(final int i, final int j,
			final double dblDelta)
	{
		data[i][j][VALUE_OBJECTS] += dblDelta;
	}

	public final double getObjectsElectrons(final int i, final int j)
	{
		return data[i][j][VALUE_OBJECTS];
	}

	public final double getTotalElectrons(final int i, final int j)
	{
		return (data[i][j][VALUE_OBJECTS] + data[i][j][VALUE_NOISES]);
	}

	/** matrice normalisee ou pas */
	public final boolean isNormalized()
	{
		return blIsNormalized;
	}

	/** l'image finale */
	public final BufferedImage getImage()
	{
		return img;
	}

}
